/*
 * xpainter.h
 * 
 * X Painter for direct XLib access. 
 * 
 * Copyright (c) 2002 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file is part of QtXLib (Fast XLib drawing interface for Qt/X11). 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */


#ifndef _INC_QTXLIB_XPAINTER_H_
#define _INC_QTXLIB_XPAINTER_H_

#include <Qt/qpaintdevice.h>
#include <Qt/qpainter.h>

//#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

//#include <X11/Xresource.h>
//#include <X11/Xatom.h>
//#include <X11/xpm.h>

namespace QTX
{

// This is a fast implementation of a painter using Xlib calls 
// instead of Qt. 
// Therefore, the function calls are NOT compatible to Qt and 
// it is NOT portable. But getting two times as fast drawing is 
// not unusual if you can use XDrawRectangles(), XDrawPoints(). 
class XPainter
{
	private:
		void _setup(const QPaintDevice *pd);
		void _cleanup();
	protected:
		Display *dpy;
		Drawable drw;
		GC gc;
		const QPaintDevice *qpd;
	public:
		// The Qt object passed is needed to get the the required 
		// parameters (display pointer, etc). 
		// Note that a pointer to the QPaintDevice is held to have 
		// access to further required data if needed. 
		// The QPainter constructor also copied the GC data using 
		// QCopyGC() (see below for info on brush & valuemask). 
		XPainter(const QPainter *qp,bool brush,ulong valuemask=0xffffffffU)
			{  _setup(qp->device());  QCopyGC(*qp,brush,valuemask);  }
		XPainter(const QPaintDevice *pd)
			{  _setup(pd);  }
		XPainter(const QPaintDevice &pd)
			{  _setup(&pd);  }
		~XPainter()
			{  _cleanup();  }
		
		// I recommend not to use them too often, but here tey are: 
		XPainter(const XPainter &src)
			{  _setup(src.qpd);  }
		XPainter &operator=(const XPainter &src)
			{  _cleanup();  _setup(src.qpd);  return(*this);  }
		
		// Of cource you may query these: 
		Display *display() const  {  return(dpy);  }
		Drawable drawable() const  {  return(drw);  }
		GC get_gc()  {  return(gc);  }
		const QPaintDevice *qpaintdev() const  {  return(qpd);  }
		
		// This is a special routine: copy the GC from a QPainter so you 
		// may use Qt routines to set the brush, etc. and then copy it. 
		// As the gc is COPIED (XCopyGC), you can safely modify the QPainter 
		// after copying without modifying the XPainter's GC. 
		// brush: true  -> copy QPainter::gc_brush;
		//        false -> copy QPainter::gc. 
		void QCopyGC(const QPainter &qb,bool brush,ulong valuemask=0xffffffffU);
		
		// Flush & Sync routines: 
		// I recommend to use XSync() regularly (not too often!)
		void XSync()		{  ::XSync(dpy,False);  }
		void XFlush()		{  ::XFlush(dpy);  }
		
		// Clear routines: 
		void ClearWindow()											{  XClearWindow(dpy,drw);  }
		void ClearArea(int x,int y,uint w,uint h,Bool exposures=0)	{  XClearArea(dpy,drw,x,y,w,h,exposures);  }
		
		// Copy from this XPainter into XPainter *to (using this gc): 
		void CopyAreaTo(XPainter *to,
			int  src_x, int  src_y,
			uint width, uint height,
			int  dest_x,int  dest_y )					{  XCopyArea(dpy,drw,to->drw,gc,src_x,src_y,width,height,dest_x,dest_y);  }
		void CopyPlaneTo(XPainter *to,
			int  src_x, int  src_y,
			uint width, uint height,
			int  dest_x,int  dest_y, ulong plane)		{  XCopyPlane(dpy,drw,to->drw,gc,src_x,src_y,width,height,dest_x,dest_y,plane);  }
		
		// Copy from XPainter *from into this XPainter (using this gc): 
		void CopyAreaFrom(XPainter *from,
			int  src_x, int  src_y,
			uint width, uint height,
			int  dest_x,int  dest_y )					{  XCopyArea(dpy,from->drw,drw,gc,src_x,src_y,width,height,dest_x,dest_y);  }
		void CopyPlaneFrom(XPainter *from,
			int  src_x, int  src_y,
			uint width, uint height,
			int  dest_x,int  dest_y, ulong plane)		{  XCopyPlane(dpy,from->drw,drw,gc,src_x,src_y,width,height,dest_x,dest_y,plane);  }
		
		// Various drawing routines: 
		void DrawArc(int x,int y,uint w,uint h,
			int angle1=0,int angle2=23040)								{  XDrawArc(dpy,drw,gc,x,y,w,h,angle1,angle2);  }
		void DrawArcs(XArc *arcs,int narcs) 						 	{  XDrawArcs(dpy,drw,gc,arcs,narcs);  }
		void DrawLine(int x1,int y1,int x2,int y2)						{  XDrawLine(dpy,drw,gc,x1,y1,x2,y2);  }
		void DrawLines(XPoint *points,int npoints,
			int mode=CoordModeOrigin)								 	{  XDrawLines(dpy,drw,gc,points,npoints,mode);  }
		void DrawPoint(int x,int y) 									{  XDrawPoint(dpy,drw,gc,x,y);  }
		void DrawPoints(XPoint *points,int npoints,
			int mode=CoordModeOrigin)									{  XDrawPoints(dpy,drw,gc,points,npoints,mode);  }
		void DrawRectangle(int x,int y,uint width,uint height)			{  XDrawRectangle(dpy,drw,gc,x,y,width,height);  }
		void DrawRectangles(XRectangle *rects,int nrects)				{  XDrawRectangles(dpy,drw,gc,rects,nrects);  }
		void DrawSegments(XSegment *segs,int nsegs) 					{  XDrawSegments(dpy,drw,gc,segs,nsegs);  }
		// Text drawing: 
		void DrawImageString(int x,int y,const char *str,int len)		{  XDrawImageString(dpy,drw,gc,x,y,str,len);  }
		void DrawString(int x,int y,const char *str,int len)			{  XDrawString(dpy,drw,gc,x,y,str,len);  }
		void DrawString(int x,int y,const char *str)					{  XDrawString(dpy,drw,gc,x,y,str,strlen(str));  }
		void DrawText(int x,int y,XTextItem *items,int nitems)			{  XDrawText(dpy,drw,gc,x,y,items,nitems);  }
		
		// Variuos fill drawing routines: 
		void FillArc(int x,int y,uint w,uint h,
			int angle1=0,int angle2=23040)								{  XFillArc(dpy,drw,gc,x,y,w,h,angle1,angle2);  }
		void FillArcs(XArc *arcs,int narcs) 							{  XFillArcs(dpy,drw,gc,arcs,narcs);  }
		void FillPolygon(XPoint *points,int npoints,
			int shape=Complex,int mode=CoordModeOrigin) 				{  XFillPolygon(dpy,drw,gc,points,npoints,shape,mode);  }
		void FillRectangle(int x,int y,uint width,uint height)			{  XFillRectangle(dpy,drw,gc,x,y,width,height);  }
		void FillRectangles(XRectangle *rects,int nrects)				{  XFillRectangles(dpy,drw,gc,rects,nrects);  }
		
		// Various settings: 
		void SetArcMode(int mode=ArcPieSlice)			{  XSetArcMode(dpy,gc,mode);  }
		void SetBackground(ulong back)					{  XSetBackground(dpy,gc,back);  }
		void SetClipMask(Pixmap cmask)					{  XSetClipMask(dpy,gc,cmask);  }
		void SetClipOrigin(int x,int y) 				{  XSetClipOrigin(dpy,gc,x,y);  }
		void SetClipRectangles(int x_origin,int y_origin,
			XRectangle *rects,int nrects,
			int ordering=Unsorted)						{  XSetClipRectangles(dpy,gc,x_origin,y_origin,rects,nrects,ordering);  }
		void SetDashes(int offset,const char *list,int n)		{  XSetDashes(dpy,gc,offset,list,n);  }
		void SetFillRule(int rule=EvenOddRule)			{  XSetFillRule(dpy,gc,rule);  }
		void SetFillStyle(int style=FillSolid)			{  XSetFillStyle(dpy,gc,style);  }
		void SetFont(Font f)							{  XSetFont(dpy,gc,f);  }
		void SetFont(XFontStruct &f)					{  XSetFont(dpy,gc,f.fid);  }
		void SetForeground(ulong fg)					{  XSetForeground(dpy,gc,fg);  }
		void SetFunction(int func=GXcopy)				{  XSetFunction(dpy,gc,func);  }
		void SetGraphicsExposures(Bool flag)			{  XSetGraphicsExposures(dpy,gc,flag);  }
		void SetLineAttributes(uint linewidth,
			int line_style=LineSolid,
			int cap_style=CapButt,
			int join_style=JoinMiter)					{  XSetLineAttributes(dpy,gc,linewidth,line_style,cap_style,join_style);  }
		void SetPlaneMask(ulong mask)					{  XSetPlaneMask(dpy,gc,mask);  }
		void SetRegion(Region r)						{  XSetRegion(dpy,gc,r);  }
		void SetState(ulong foreg,ulong backg,
			int func,ulong planemask)					{  XSetState(dpy,gc,foreg,backg,func,planemask);  }
		void SetStipple(Pixmap stipple) 				{  XSetStipple(dpy,gc,stipple);  }
		void SetSubwindowMode(int mode=ClipByChildren)	{  XSetSubwindowMode(dpy,gc,mode);  }
		void SetTile(Pixmap tile)						{  XSetTile(dpy,gc,tile);  }
		void SetTSOrigin(int x,int y)					{  XSetTSOrigin(dpy,gc,x,y);  }
		
		// Some query functions: 
		Status QueryBestSize(int cclass,
			uint width,uint height,
			uint *width_return,uint *height_return) 	{  return(XQueryBestSize(dpy,cclass,drw,width,height,width_return,height_return));  }
		Status QueryBestStipple(
			uint width,uint height,
			uint *width_return,uint *height_return) 	{  return(XQueryBestStipple(dpy,drw,width,height,width_return,height_return));  }
		Status QueryBestTile(
			uint width,uint height,
			uint *width_return,uint *height_return) 	{  return(XQueryBestTile(dpy,drw,width,height,width_return,height_return));  }
};

}  // end of namespace QTX

#endif  /* _INC_QTXLIB_XPAINTER_H_ */
