/*
 * area.cc
 * 
 * BURN simulation core implementation. 
 * 
 * Copyright (c) 2002--2004 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include <stdio.h>
#include <assert.h>
#include <string.h>

#include "burncore.h"

// TUNING OPTIONS: 
// On an AthlonXP with gcc-3.0.5, these two give about 40% time decrease: 
#define NEIGH_CENTER 1   /* extra code for center cells */
#define NEIGH_UNROLL 1   /* unrolled neighbour check loop (requires NEIGH_CENTER) */

// This decreases speed for me, so don't use: 
#define EXTRA_BPF_MEMSET 0  /* use memset() to clear bpf_accu, don't do in loop */


// CHECK OPTIONS: (Switch them all off for performant operation.) 
#define CHECK_COUNTER 0


inline void BurnArea::_NeighboursAddBurnFlags_Center(int x,int y)
{
#if NEIGH_CENTER
	// Now, for all neighbours *snc, add the burn propabilities: 
	// This is done by setting the bit in pbf_accu[this_cell] 
	// describing that *sc is a burning neighbour of the (neighbour) 
	// cell *snc; this info is then used in the second loop to 
	// set it on fire if needed. 
	
	#if NEIGH_UNROLL
	
	// *** UNROLLED VERSION ***
	size_t idx=arrayidx(x-1,y-1);
	BCell *nc=&cellarray[idx];
	unsigned char *bpf=&pbf_accu[idx];
	_CheckCell(nc,bpf,'\x01');	 ++nc;  ++bpf;
	_CheckCell(nc,bpf,'\x02');	 ++nc;  ++bpf;
	_CheckCell(nc,bpf,'\x04');	 nc+=width-2;	 bpf+=width-2;
	_CheckCell(nc,bpf,'\x08');	 nc+=2; 	 bpf+=2;
	_CheckCell(nc,bpf,'\x10');	 nc+=width-2;	 bpf+=width-2;
	_CheckCell(nc,bpf,'\x20');	 ++nc;  ++bpf;
	_CheckCell(nc,bpf,'\x40');	 ++nc;  ++bpf;
	_CheckCell(nc,bpf,'\x80');
	//assert(nc==&cellarray[arrayidx(x+1,y+1)]);
	
	#else  /* !NEIGH_UNROLL */
	
	// *** LOOP VERSION ***
	uint mask=1;
	for(int dy=-1; dy<=1; dy++)
	{
		for(int dx=-1; dx<=1; dx++)
		{
			if(!dx && !dy)  continue;  // NOT: goto cont;
			// Jump (goto) block: 
			{
				int nx=x+dx;
				int ny=y+dy;
				
				// Okay, neighbour at nx,ny: 
				BCell *nc=cell(nx,ny);
				if(nc->state!=BCell::Living)  goto cont;
				
				// Queue it in the first_flame equeue if not 
				// already queued: 
				unsigned char *acc=bpbfaccu(nx,ny);
				if(!(*acc))
				{  _FLAMEAddFirst(nc);  }
				
				// Okay, add flag to (burn prop flag) accumulator: 
				(*acc)|=(unsigned char)mask;
			}
			cont: mask<<=1;
		}
	}
	
	#endif  /* NEIGH_UNROLL */
#endif  /* NEIGH_CENTER */
}


inline void BurnArea::_NeighboursAddBurnFlags_Border(int x,int y)
{
	// The same as _NeighboursAddBurnFlags_Center but with border check: 
	uint mask=1;
	for(int dy=-1; dy<=1; dy++)
	{
		for(int dx=-1; dx<=1; dx++)
		{
			if(!dx && !dy)  continue;  // NOT: goto cont;
			// Jump (goto) block: 
			{
				int nx=x+dx;
				if(nx<0 || nx>=width)
				{
					if(!torus)  goto cont;
					nx=(nx+width)%width;
				}
				int ny=y+dy;
				if(ny<0 || ny>=height)
				{
					if(!torus)  goto cont;
					ny=(ny+height)%height;
				}

				// Okay, neighbour at nx,ny: 
				BCell *nc=cell(nx,ny);
				if(nc->state!=BCell::Living)  goto cont;
				
				// Queue it in the first_flame equeue if not 
				// already queued: 
				unsigned char *acc=bpbfaccu(nx,ny);
				if(!(*acc))
				{  _FLAMEAddFirst(nc);  }
				
				// Okay, add flag to (burn prop flag) accumulator: 
				(*acc)|=(unsigned char)mask;
			}
			cont: mask<<=1;
		}
	}
}


// Only call for cells which are in state Burning: 
inline int BurnArea::_CellTick(BCell *sc,BCell *prev_sc)
{
	//assert(sc->state==BCell::Burning);
	
	size_t idx=sc-cellarray;
	
	int x=idx2xco(idx);
	int y=idx2yco(idx);
	
	#if NEIGH_CENTER
	if(!x || !y || x+1==width || y+1==height)
	{  _NeighboursAddBurnFlags_Border(x,y);  }
	else
	{  _NeighboursAddBurnFlags_Center(x,y);  }
	#else  /* !NEIGH_CENTER */
	_NeighboursAddBurnFlags_Border(x,y);
	#endif
	
	if(!(--sc->mass))
	{
		// Burning -> Dead transition: 
		sc->state=BCell::Dead;
		_BRNDelete(sc,prev_sc);
		_StateChangeUpdCnt(idx,BCell::Burning,BCell::Dead);
		
		return(0);
	}
	
	return(1);
}


inline void BurnArea::_CellTickFire(BCell *sc)
{
	size_t idx=sc-cellarray;
	unsigned char *acc=&pbf_accu[idx];
	
	int prop=burn_prop_table[int(*acc)];
	if(prop && rnd.decide(prop))
	{
		// Set cell on fire: 
		// Living -> Burning transition: 
		sc->state=BCell::Burning;
		_BRNAddFirst(sc);
		_StateChangeUpdCnt(idx,BCell::Living,BCell::Burning);
	}
	
	#if !EXTRA_BPF_MEMSET
	// Inportant: reset accumulator for next round. 
	*acc='\0';
	#endif
}


// Simulate one sim tick. 
// Return value: 
//  1 -> sim not in stable state 
//  0 -> now stable state, next tick will not change anything 
int BurnArea::SimTick()
{
	// First round: mass loss and cell dying: 
	// For all cells in Burning state: 
	BCell *prev_sc=NULL;
	for(BCell *sc=first_brn; sc; sc=sc->next)
	{
		if(_CellTick(sc,prev_sc))
		{  prev_sc=sc;  }
	}
	
	// Second round: set cells on fire: 
	for(BCell *sc=first_flame; sc; )
	{
		BCell *tmp=sc;
		sc=sc->next;
		_CellTickFire(tmp);
	}
	first_flame=NULL;
	
	#if EXTRA_BPF_MEMSET
	memset(pbf_accu,0,width*height);
	#endif
	
	#if CHECK_COUNTER
	#warning CHECK_COUNTER active. 
	size_t cnt[BCell::_LAST];
	memset(cnt,0,sizeof(cnt));
	for(int y=0; y<height; y++)
	{
		for(int x=0; x<width; x++)
		{  ++cnt[cell(x,y)->state];  }
	}
	for(int i=0; i<BCell::_LAST; i++)
	{
		if(cnt[i]==ncell[i])  continue;
		fprintf(stderr,"MISMATCH: state=%d, is=%d, shall=%d\n",
			i,ncell[i],cnt[i]);
	}
	#endif  /* CHECK_COUNTER */
	
	return(ncell[BCell::Burning] ? 1 : 0);
}


int BurnArea::SetFire(int x,int y)
{
	if(x<0 || x>=width || y<0 || y>=width)
	{  return(2);  }
	
	switch(cell(x,y)->state)
	{
		case BCell::Living:
		{
			size_t idx=arrayidx(x,y);
			// Set cell on fire: Living -> Burning transition: 
			BCell *sc=&cellarray[idx];
			sc->state=BCell::Burning;
			_BRNAddFirst(sc);
			_StateChangeUpdCnt(idx,BCell::Living,BCell::Burning);
		}  return(0);
		case BCell::Burning:
			return(1);
	}
	return(3);
}


void BurnArea::RandomFill(double tree_prop,int tree_mass)
{
	if(tree_mass<1)  tree_mass=1;
	
	_ResetCounters();
	first_brn=NULL;
	first_flame=NULL;
	int t_prop=RandomGenerator::PropD2I(tree_prop);
	
	size_t ncells=size_t(width)*size_t(height);
	for(BCell *i=cellarray,*cae=i+ncells; i<cae; i++)
	{
		if(rnd.decide(t_prop))
		{
			i->state=BCell::Living;
			i->mass=tree_mass;
		}
		else
		{
			i->state=BCell::Empty;
			i->mass=0;
		}
		i->next=NULL;
		++ncell[i->state];
	}
}


// Sets up the array to be completely BCell::Empty, mass=0. 
int BurnArea::SetSize(int w,int h,int _torus)
{
	_CleanupArea();
	
	if(!w)  w=1;
	if(!h)  h=1;
	
	width=w;
	height=h;
	torus=_torus;
	
	//fprintf(stderr,"TORUS=%d\n",torus);
	
	// Set up cell array: 
	size_t ncells=size_t(width)*size_t(height);
	cellarray=(BCell*)malloc(ncells*sizeof(BCell));
	assert(cellarray);
	
	// Fill with empty cells: 
	for(BCell *i=cellarray,*cae=i+ncells; i<cae; i++)
	{
		i->state=BCell::Empty;
		i->mass=0;
		i->next=NULL;
	}
	ncell[BCell::Empty]=ncells;
	
	// Set up burn prop flags accumulator: 
	pbf_accu=(unsigned char *)malloc(ncells);
	for(size_t i=0; i<ncells; i++)
	{  pbf_accu[i]='\0';  }
	
	#if USE_STATE_CHG
	// Set up state change flags: 
	state_chg=(unsigned char *)malloc(ncells);
	for(size_t i=0; i<ncells; i++)
	{  state_chg[i]='\0';  }
	#endif
	
	return(0);
}


void BurnArea::GetBurnProp(double *dest)
{
	for(double *s=&burn_prop[0][0]; s<=&burn_prop[2][2]; s++)
	{  *(dest++)=*s;  }
}


int BurnArea::SetBurnProp(const double *prop)
{
	// Copy burn props: 
	for(double *d=&burn_prop[0][0]; d<=&burn_prop[2][2]; d++)
	{  *d=*(prop++);  }
	// Force sane prop for cell itself: 
	burn_prop[1][1]=0;
	
	// Okay, calculate burn prop table: 
	double bpt[256];
	for(uint i=0; i<256; i++)
	{
		// Neighbour bit meaning: 
		//  0 1 2 
		//  3 # 4       NOTE also: burn_prop[y][x]
		//  4 5 7 
		double p=1.0;
		uint mask=1;
		for(int ny=0; ny<3; ny++)
		{
			for(int nx=0; nx<3; nx++)
			{
				if(nx==1 && ny==1)  continue;
				if(i & mask)
				{  p*=(1.0-burn_prop[ny][nx]);  }
				mask<<=1;
			}
		}
		bpt[i]=1.0-p;
		
		assert(p>=0.0 && p<=1.0);
	}
	
	// Set up integer burn prop table: 
	for(int i=0; i<256; i++)
	{  burn_prop_table[i]=RandomGenerator::PropD2I(bpt[i]);  }
	
	return(0);
}


void BurnArea::_ResetCounters()
{
	for(int i=0; i<BCell::_LAST; i++)
	{  ncell[i]=0;  }
}


void BurnArea::_CleanupArea()
{
	if(cellarray)   free(cellarray);   cellarray=NULL;
	if(pbf_accu)    free(pbf_accu);    pbf_accu=NULL;
	#if USE_STATE_CHG
	if(state_chg)   free(state_chg);   state_chg=NULL;
	#endif
	first_brn=NULL;
	first_flame=NULL;
	width=height=0;
	_ResetCounters();
}


BurnArea::BurnArea() : 
	rnd()
{
	cellarray=NULL;
	first_brn=NULL;
	first_flame=NULL;
	pbf_accu=NULL;
	#if USE_STATE_CHG
	state_chg=NULL;
	#else
	update_callback=NULL;
	callback_dptr=NULL;
	#endif
	
	width=0;
	height=0;
	torus=0;
	
	_ResetCounters();
	
	// Set up defaults: 
	double bp[3][3]={
		{ 0.5, 0.5, 0.5 },
		{ 0.5, 0.0, 0.5 },
		{ 0.5, 0.5, 0.5 } };
	SetBurnProp(bp[0]);
	
	SetSize(100,100,0);
}

BurnArea::~BurnArea()
{
	_CleanupArea();
}
