/*
 * areawindow.cc
 * 
 * Qxburn's simulation area window. 
 * 
 * Copyright (c) 2002--2003 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include "areawindow.h"

#include <Qt/qpushbutton.h>
#include <Qt/qcheckbox.h>
#include <Qt/qlayout.h>
#include <Qt/qlabel.h>
#include <Qt/qvalidator.h>

#include "../gui/ignpars.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>


const int updates_per_timer=3;


void AreaWindow::resizeEvent(QResizeEvent *)
{
	int nw=size().width();
	int nh=size().height();
	
	if(nw<int(area.AreaWidth()) || nh<int(area.AreaHeight()))
	{
		fprintf(stderr,"OOps: window smaller than min size?!\n");
		QSize minsize(area.AreaWidth(),area.AreaHeight());
		setMinimumSize(minsize);
	}
	
	nw=size().width();
	nh=size().height();
	
	cellw=nw/area.AreaWidth();   if(!cellw) cellw=1;
	cellh=nh/area.AreaHeight();  if(!cellh) cellh=1;
	
	cellx0=(nw - cellw*area.AreaWidth())/2;
	celly0=(nh - cellh*area.AreaHeight())/2;
	
	bool old_val=use_ppnt;
	use_ppnt=(cellw==1 && cellh==1);
	if(old_val!=use_ppnt)
	{
		rpnt.Flush();
		ppnt.Flush();
	}
	//fprintf(stderr,"DRAWING=%s\n",use_ppnt ? "point" : "rect");
}


void AreaWindow::paintEvent(QPaintEvent *ev)
{
	if(!isVisible())  return;
	
	QRect qr=ev->rect();
	int px=qr.x();
	int py=qr.y();
	int pw=qr.width();
	int ph=qr.height();
	
	int sw=area.AreaWidth();
	int sh=area.AreaHeight();
	
	int x0=(px-cellx0)/int(cellw);  if(x0<0)   x0=0;
	int y0=(py-celly0)/int(cellh);  if(y0<0)   y0=0;
	int xe=x0+2+pw/int(cellw);      if(xe>sw)  xe=sw;
	int ye=y0+2+ph/int(cellh);      if(ye>sh)  ye=sh;
	
	//fprintf(stderr,">%d %d %d %d<\n",x0,y0,xe,ye);
	for(int y=y0; y<ye; y++)
	{
		for(int x=x0; x<xe; x++)
		{
			const BCell *bc=area.AreaCell(x,y);
			if(use_ppnt)
			{  ppnt.AddPoint(bc->state,x+cellx0,y+celly0);  }
			else
			{  rpnt.AddRect(bc->state,x*cellw+cellx0,y*cellh+celly0,cellw,cellh);  }
		}
	}
	
	if(use_ppnt)  ppnt.Flush();
	else  rpnt.Flush();
	XSync(rpnt[0]->display(),False);
}


void AreaWindow::DoAutoIgnite(IgniteParsDialog *ignpars)
{
	ignpars->DoAutoIgnite(&area);
	
	// After setting fire, the painter must be flushed: 
	if(use_ppnt)  ppnt.Flush();
	else  rpnt.Flush();
	_UpdateStat();
}


void AreaWindow::mouseMoveEvent(QMouseEvent *ev)
{
	if(!(ev->state()&LeftButton) && !(ev->button()&LeftButton))
	{  return;  }
	
	int mx=ev->x(),my=ev->y();
	int cx = (mx-cellx0)/int(cellw);
	int cy = (my-celly0)/int(cellh);
	if(cx<0 || cy<0 || 
	   cx>=int(area.AreaWidth()) || cy>=int(area.AreaHeight()))
	{  return;  }
	
	// Mouse cursor pressed and over cell cx,cy. 
	// Set fire on cell: 
	if(area.SetFire(cx,cy)==0)
	{
		// After setting fire, the painter must be flushed: 
		if(use_ppnt)  ppnt.Flush();
		else  rpnt.Flush();
		_UpdateStat();
	}
}

void AreaWindow::mousePressEvent(QMouseEvent *ev)
{  mouseMoveEvent(ev);  }


void AreaWindow::keyPressEvent(QKeyEvent *ev)
{
	emit key_press_event(ev);
}

void AreaWindow::keyReleaseEvent(QKeyEvent *ev)
{
	emit key_release_event(ev);
}


void AreaWindow::timerEvent(QTimerEvent *)
{
	assert(running);
	
	for(int i=updates_per_timer; i>=0; i--)
	{
		int rv=area.SimTick();
		if(rv)
		{  ++tick_cnt;  }
		
		#if USE_STATE_CHG
		_UpdateChangedCells();
		#endif
		
		if(use_ppnt)  ppnt.Flush();
		else  rpnt.Flush();
		XSync(rpnt[0]->display(),False);
		
		if(iterations_left>0)
		{  --iterations_left;  }
		
		if(!rv || !iterations_left)
		{
			if(!rv)  no_more_trees=1;
			StopSim();
			break;
		}
		
		// Only do one tick per timer if a delay is set: 
		if(delay_val)
		{  break;  }
	}
	
	if(delay_mode==1 && running)  // ticks per delay
	{
		assert(delay_val);
		
		++ticks_since_delay;
		if(ticks_since_delay>=delay_val)
		{
			killTimers();
			startTimer(9);  // 9msec
			ticks_since_delay=0;
		}
		else if(ticks_since_delay==1)
		{
			killTimers();
			startTimer(0);
		}
	}
	
	if(running && 
	   (delay_val || !(tick_cnt%25)) )
	{  _UpdateStat();  }
}


void AreaWindow::SetDelay(int dmode,int dval)
{
	delay_val=(dval<0) ? 0 : dval;
	delay_mode=delay_val ? dmode : 0;
	// delay_mode ALWAYS 0 if delay_val is 0. 
	
	if(running)
	{
		killTimers();
		if(delay_mode==1)  // ticks per delay
		{
			startTimer(0);
		}
		else if(delay_mode==0)  // delay per tick
		{
			ticks_since_delay=0;
			startTimer(delay_val);
		}
		else  assert(0);
	}
}


void AreaWindow::ChangeBurnProp(const double *prop)
{
	area.SetBurnProp(prop);
}


void AreaWindow::ContSim(int iteration_limit)
{
	if(running)  return;
	
	if(delay_mode==0)  // msec delay per tick
	{  startTimer(delay_val);  }
	else if(delay_mode==1) // ticks per delay
	{  startTimer(0);  ticks_since_delay=0;  }
	else  assert(0);
	
	running=1;
	iterations_left=iteration_limit ? iteration_limit : (-1);
}

void AreaWindow::StopSim()
{
	if(!running)  return;
	
	killTimers();
	running=0;
	
	int reason=0;
	if(!iterations_left)  reason=2;
	else if(no_more_trees)  reason=1;
	emit sim_has_stopped(reason);
	
	_UpdateStat();
}


void AreaWindow::fillin_area_setup(AreaSetupPars *spar)
{
	spar->width=area.AreaWidth();
	spar->height=area.AreaHeight();
	spar->torus_area=area.AreaIsTorus();
	spar->valid=true;
}

void AreaWindow::fillin_burnprop(double *dest)
{
	area.GetBurnProp(dest);
}


void AreaWindow::fillin_sim_setup(SimSetupPars *spar)
{
	*spar=ssp;
	spar->valid=true;
}


void AreaWindow::area_setup_changed(AreaSetupPars *spar)
{
	area.SetSize(spar->width,spar->height,spar->torus_area);
	setMinimumSize(QSize(area.AreaWidth(),area.AreaHeight()));
	erase();
	resizeEvent(NULL);
	//area.RandomFill(ssp.tree_prop,ssp.tree_mass);
	//Update();
	//ClearArea() gets now called. 
}

void AreaWindow::ClearArea(SimSetupPars *_ssp)
{
	ssp=*_ssp;
	if(ssp.tree_prop<0.0)  ssp.tree_prop=0.0;
	if(ssp.tree_prop>1.0)  ssp.tree_prop=1.0;
	if(ssp.tree_mass<1)  ssp.tree_mass=1;
	
	area.RandomFill(ssp.tree_prop,ssp.tree_mass);
	no_more_trees=0;
	tick_cnt=0;
	_UpdateStat();
	
	Update();
}


// Re-draw complete screen: 
void AreaWindow::Update()
{
	//fprintf(stderr,"COMPLETE UPDATE\n");
	
	// Complete update: 
	int sh=area.AreaHeight();
	int sw=area.AreaWidth();
	for(int y=0; y<sh; y++)
	{
		for(int x=0; x<sw; x++)
		{
			const BCell *bc=area.AreaCell(x,y);
			if(use_ppnt)
			{  ppnt.AddPoint(bc->state,x+cellx0,y+celly0);  }
			else
			{  rpnt.AddRect(bc->state,x*cellw+cellx0,y*cellh+celly0,cellw,cellh);  }
		}
	}
	
	if(use_ppnt)  ppnt.Flush();
	else  rpnt.Flush();
	XSync(rpnt[0]->display(),False);
}


#if USE_STATE_CHG
// Only draw cells which chagned state: 
void AreaWindow::_UpdateChangedCells()
{
	int sh=area.AreaHeight();
	int sw=area.AreaWidth();
	for(int y=0; y<sh; y++)
	{
		for(int x=0; x<sw; x++)
		{
			if(area.StateChangedReset(x,y))
			{
				const BCell *bc=area.AreaCell(x,y);
				if(use_ppnt)
				{  ppnt.AddPoint(bc->state,x+cellx0,y+celly0);  }
				else
				{  rpnt.AddRect(bc->state,x*cellw+cellx0,y*cellh+celly0,cellw,cellh);  }
			}
		}
	}
	
	// flush done outside. 
}
#else  /* !USE_STATE_CHG */
inline void AreaWindow::_UpdateCell(int x,int y,BCell *bc)
{
	if(use_ppnt)
	{  ppnt.AddPoint(bc->state,x+cellx0,y+celly0);  }
	else
	{  rpnt.AddRect(bc->state,x*cellw+cellx0,y*cellh+celly0,cellw,cellh);  }
}

// Simple wrapper: 
void _UpdateHandler(void *thisptr,int x,int y,BCell *c)
{
	((AreaWindow*)thisptr)->_UpdateCell(x,y,c);
}
#endif


void AreaWindow::Map()
{
	if(isVisible())
	{
		hide();
		#warning disable sim updates?
	}
	else
	{
		setMinimumSize(QSize(area.AreaWidth(),area.AreaHeight()));
		show();
	}
}


AreaWindow::AreaWindow(QWidget *_parent,const char *name,WFlags f) : 
	QWidget(_parent,name,f),
	area(),
	rpnt(BCell::_LAST),
	ppnt(BCell::_LAST)
{
	#if !USE_STATE_CHG
	area.RegisterUpdateCallback(&_UpdateHandler,this);
	#endif
	
	cellx0=0;
	celly0=0;
	cellw=3;
	cellh=3;
	use_ppnt=0;
	
	running=0;
	area.SetSize(300,300,/*torus_area=*/0);
	ssp.tree_prop=0.5;
	ssp.tree_mass=5;
	
	delay_mode=0;
	delay_val=9;
	ticks_since_delay=0;
	
	iterations_left=-1;
	
	no_more_trees=1;
	
	tick_cnt=0;
	statwin=NULL;
	
	QPainter pnt(this);
	
	// Empty: 
	pnt.setBrush(QBrush(Qt::white,Qt::SolidPattern));
	rpnt.Setup(BCell::Empty,&pnt,1024);
	ppnt.Setup(BCell::Empty,&pnt,2048);
	setBackgroundColor(Qt::white);
	
	// Living: 
	pnt.setBrush(QBrush(Qt::green,Qt::SolidPattern));
	rpnt.Setup(BCell::Living,&pnt,1024);
	ppnt.Setup(BCell::Living,&pnt,2048);
	
	// Burning: 
	pnt.setBrush(QBrush(Qt::red,Qt::SolidPattern));
	rpnt.Setup(BCell::Burning,&pnt,1024);
	ppnt.Setup(BCell::Burning,&pnt,2048);
	
	// Dead: 
	pnt.setBrush(QBrush(Qt::black,Qt::SolidPattern));
	rpnt.Setup(BCell::Dead,&pnt,1024);
	ppnt.Setup(BCell::Dead,&pnt,2048);
	
	// Correct size: 
	resize(area.AreaWidth()*cellw,area.AreaHeight()*cellh);
	
	setCaption("Burn Simulator");
}


AreaWindow::~AreaWindow()
{
}
