/*
 * mainwindow.cc
 * 
 * Qxburn's main window. 
 * 
 * Copyright (c) 2002--2006 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#include "config.h"

#include "mainwindow.h"

#include "../gui/ignpars.h"
#include "../gui/burnprop.h"
#include "../gui/areatree.h"
#include "../gui/areasetup.h"
#include "statdiag.h"

#include <Qt/qpushbutton.h>
#include <Qt/qcheckbox.h>
#include <Qt/qlayout.h>
#include <Qt/qlabel.h>
#include <Qt/qlineedit.h>
#include <Qt/qvalidator.h>
#include <Qt/qspinbox.h>
#include <Qt/qgroupbox.h>
#include <Qt/qmessagebox.h>
#include <Qt/qtabwidget.h>
#include <Qt/qtooltip.h>
#include <Qt/qcombobox.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <assert.h>


void MainWindow::quit_clicked()
{
	close();
}

void MainWindow::info_clicked()
{
	QMessageBox::about(this,
		"Burn Simulator / Statistics",
		"<center><b>ST Burn</b>, version " VERSION "</center>"
		"Fast burn simulator / statistics<br>(c) 2002--2006 by Wolfgang Wieser.<br><br>"
		"This program may be copied unter the terms of the GNU GPL.<br>");
}


void MainWindow::reset_clicked()
{
	stat_diag->Reset();
	
	sweep_step=0;
	sweep_iter=0;
	
	// This is normally not needed becuase of slot sim_setup_pars_changed(). 
	SimSetupPars ssp;  ssp.valid=false;
	atd->ReadValues(&ssp);
	assert(ssp.valid);
	
	if(!stat_diag->isVisible())
	{  stat_diag->DoShow();  }
	
	_UpdateValues();
	
	stop_cont_button->setText("&Start");
}


void MainWindow::stop_cont_clicked()
{
	if(running)
	{  _StopSim();  }
	else
	{  _ContSim();  }
}

void MainWindow::_StopSim()
{
	if(!running)  return;
	stop_cont_button->setText("Cont");
	stop_cont_button->setAccel(ALT + Key_S);
	
	stat_diag->ReDraw();
	
	running=0;
	sps_time_valid=0;
	killTimers();
}

void MainWindow::_ContSim()
{
	if(running)  return;
	stop_cont_button->setText("&Stop");
	
	startTimer(0);
	running=1;
}


void MainWindow::sim_setup_pars_request_slot(SimSetupPars *spar)
{
	*spar=ssp;
	spar->valid=true;
}


void MainWindow::sim_setup_pars_changed()
{
	// Get new values: 
	atd->ReadValues(&ssp);
	atd->Update();
}


void MainWindow::area_setup_changed(AreaSetupPars *spar)
{
	assert(spar->valid);
	
	sps_time_valid=0;
	
	int rv=area.SetSize(spar->width,spar->height,spar->torus_area);
	assert(!rv);
	stat_diag->area_setup_changed(spar);
	
	if(!stat_diag->isVisible())
	{  stat_diag->DoShow();  }
	
	_UpdateValues();
}


void MainWindow::burn_prop_change(double *bprop)
{
	area.SetBurnProp(bprop);
	//_UpdateValues();
}

void MainWindow::burn_prop_request_slot(double *bprop)
{
	area.GetBurnProp(bprop);
}

void MainWindow::fillin_area_setup(AreaSetupPars *spar)
{
	spar->width=area.AreaWidth();
	spar->height=area.AreaHeight();
	spar->torus_area=area.AreaIsTorus();
	spar->valid=true;
}


void MainWindow::diag_limit_changed(int new_limit)
{
	stat_diag->ChangeEntryLimit(new_limit);
}


// stat: +1 -> press; -1 -> release
void MainWindow::_HandleKeyEvent(QKeyEvent *ev,int stat)
{
	if(ev->isAutoRepeat() || stat==+1)
	{  ev->ignore();  return;  }
	
	ev->ignore();/*switch(ev->key())
	{
		case Key_Return:
		case Key_Enter:
			// Let's see where we are: 
			if(atd->YouHaveFocus())
			{
				reset_button->animateClick();
				stop_cont_button->setFocus();
			}
			else if(!areawin.HasBurningTrees())
			{
				//reset_button->animateClick();
				atd->YouGetFocus(0);
			}
			else
			{
				stop_cont_button->animateClick();
			}
			break;
		case Key_A:  area_button->animateClick();  break;
		case Key_I:  ignite_button->animateClick();  break;
		case Key_R:  reset_button->animateClick();  break;
		case Key_S:  stop_cont_button->animateClick();  break;
		case Key_T:  atd->YouGetFocus(0);  break;
		case Key_M:  atd->YouGetFocus(1);  break;
		default:  ev->ignore();  return;
	}*/
}

void MainWindow::keyPressEvent(QKeyEvent *ev)
{
	_HandleKeyEvent(ev,+1);
}

void MainWindow::keyReleaseEvent(QKeyEvent *ev)
{
	_HandleKeyEvent(ev,-1);
}


void MainWindow::_UpdateValues()
{
	atd->Update();
	bpropd->Update();
}


double MainWindow::_GetTreeProp()
{
	if(sweep_step==0)
	{
		// Initial step: seep along tree prop from 0 to 1, 
		// continuously 250 entries. 
		double prop=sweep_iter/249.0;
		++sweep_iter;
		if(sweep_iter>=250)
		{  sweep_step=1;  }
		return(prop);
	}
	else if(sweep_step==1 || sweep_step==3)
	{
		StatisticsDiagram::DivEntry *darray;
		int nent;
		if(sweep_step==1)
		{  stat_diag->GetDivArrayA(&nent,&darray);  }
		else
		{  stat_diag->GetDivArrayB(&nent,&darray);  }
		int max_sigma_i=-1;
		double max_sigma=-1.0;
		for(int i=0; i<nent; i++)
		{
			StatisticsDiagram::DivEntry *de=&darray[i];
			double avg=de->sum_dead_frac/de->nent;
			double sigma=de->sq_sum_dead_frac/de->nent-avg*avg;
			//fprintf(stderr,"%.2f..%.2f: nent=%u; avg=%f; sigma=%f\n",
			//	de->min_tree_prop,de->max_tree_prop,
			//	de->nent,
			//	avg,sigma);
			if(max_sigma<sigma)
			{  max_sigma=sigma;  max_sigma_i=i;  }
		}
		if(sweep_step==1)
		{
			// Center of DivEntry with largest sigma. 
			sweep2_center=
				(darray[max_sigma_i].min_tree_prop+
				 darray[max_sigma_i].max_tree_prop) / 2.0;
			sweep2_delta=darray[max_sigma_i].max_tree_prop-darray[max_sigma_i].min_tree_prop;
			sweep2_delta*=3.0;  // For the neighbours. 
		}
		else
		{
			// Center: weighted middle of 30% in the center. 
			int i0=max_sigma_i-(nent+3)/6;
			int i1=max_sigma_i+(nent+3)/6;
			if(i0<1)  i0=0;
			if(i1>=nent-1)  i1=nent-2;
			double sigma_sum=0.0;
			double sigma_center=0.0;
			double avgdelta_sum=0.0;
			double avgdelta_center=0.0;
			for(int i=i0; i<=i1; i++)
			{
				StatisticsDiagram::DivEntry *de=&darray[i];
				double avg0=de->sum_dead_frac/de->nent;
				double avgp=(de-1)->sum_dead_frac/(de-1)->nent;
				double avgn=(de+1)->sum_dead_frac/(de+1)->nent;
				// This could be expressed simplier...
				double avg_b = (avg0-avgp)/2.0;
				double avg_t = (avgn-avg0)/2.0;
				double delta_avg = avg_t-avg_b;
				//fprintf(stderr,"d=%f\n",delta_avg);
				double sigma=de->sq_sum_dead_frac/de->nent-avg0*avg0;
				sigma_sum+=sigma;
				sigma_center+=sigma*(de->min_tree_prop+de->max_tree_prop)/2.0;
				avgdelta_sum+=delta_avg;
				avgdelta_center+=delta_avg*(de->min_tree_prop+de->max_tree_prop)/2.0;
			}
			sigma_center/=sigma_sum;
			avgdelta_center/=avgdelta_sum;
			//fprintf(stderr,"Sigma center=%f; AvgDelta center=%f <--- garbage!!!\n",
			//	sigma_center,avgdelta_center);
				
			sweep2_center=sigma_center*0.97;
			// Protection...
			if(!finite(sweep2_center) || sweep2_center<=0.0 || sweep2_center>=1.0)
			{  sweep2_center=0.5;  }
			
			//fprintf(stderr,"Choosing %f\n",sweep2_center);
		}
		
		sweep_iter=0;
		
		++sweep_step;
		// NO return.
	}
	// NO else!
	if(sweep_step==2)
	{
		// Second step: seep along tree prop for the DivEntry with 
		// the largest sigma and its neighbours. 
		// Continuously 500 entries. 
		double prop;
		// This is stupid solution... FIXME:
		do
		{
			prop=(sweep_iter/490.0-0.5)*sweep2_delta+sweep2_center;
			++sweep_iter;
		}
		while(prop<0.0);
		if(sweep_iter>=500)
		{  sweep_step=3;  }
		if(prop>=1.0)
		{  sweep_step=3;  prop=1.0;  }
		return(prop);
	}
	else if(sweep_step==4)
	{
		return(double(rndgen.get_random())/double(RandomMax-1));
		#warning No clever algorithm...
		#if 0
		#warning OPTIMIZE ME!
		for(;;)
		{
			double r11=double(rndgen.get_random()-RandomMax/2)/(RandomMax/2);  // -1..+1
			double rsgn;
			if(r11<0.0)
			{  rsgn=-1.0;  r11=-r11;  }
			else
			{  rsgn=1.0;  }
			double p=pow(r11,1.5)*rsgn+sweep2_center;
			if(p>=0.0 && p<=1.0)
			{  return(p);  }
		}
		#endif
	}
	assert(0);
}


// Copied from HLib (c) Wolfgang Wieser...
static inline long msec_elapsed(const struct timeval *old)
{
	timeval current;
	gettimeofday(&current,NULL);
	return(
		(current.tv_sec  - old->tv_sec )*1000L + 
		(current.tv_usec - old->tv_usec)/1000L );
}
static inline long usec_elapsed(const struct timeval *old,
	const struct timeval *current)
{
	return(
		(current->tv_sec  - old->tv_sec )*1000000L + 
		(current->tv_usec - old->tv_usec) );
}

void MainWindow::timerEvent(QTimerEvent *)
{
	#warning "************* CALL srandom/srand FROM TIME TO TIME... (also QXBurn) *************"
	#warning "** Special buttons: re-calc (gegen Rundungsfehler), curve ein/aus, Punkte ein/aus, sigma ein/aus **"
	#warning "*********************************************************************************"
	
	if(running)
	{
		// Do some more sim ticks: 
		
		if(!area.HasBurningCells())
		{
			if(!sps_time_valid)
			{
				gettimeofday(&sps_starttime,NULL);
				sps_time_valid=1;
			}
			
			ssp.tree_prop=_GetTreeProp();
			
			area.RandomFill(ssp.tree_prop,ssp.tree_mass);
			ignitepars->DoAutoIgnite(&area);
			iterations=0;
		}
		
		timeval starttime;
		gettimeofday(&starttime,NULL);
		for(;;)
		{
			for(int cnt=25; cnt; cnt--)
			{
				int rv=area.SimTick();
				++iterations;
				if(!rv)
				{
					long usec=-1;
					if(sps_time_valid)
					{
						timeval current;
						gettimeofday(&current,NULL);
						usec=usec_elapsed(&sps_starttime,&current);
						sps_starttime=current;
					}
					stat_diag->AddRecord(ssp.tree_prop,&area,iterations,usec);
					goto out;
				}
			}
			if(msec_elapsed(&starttime)>30)  break;
		}
		out:;
	}
	else
	{
		/* only for setup */
		killTimers();
		
		area.SetSize(100,100,0);
		
		_UpdateValues();
		// Make sure burn props are according to wind params: 
		bpropd->windparam_change();
		
		diag_limit->setValue(stat_diag->GetEntryLimit());
		
		show();
		
		reset_clicked();
		
		stop_cont_button->setFocus();
	}
}

MainWindow::MainWindow(QWidget *_parent,const char *name,WFlags f) : 
	QWidget(_parent,name,f),
	asetup(NULL,NULL,WStyle_Customize | WStyle_DialogBorder),   // no parent
	rndgen()
{
	iterations=0;
	
	ssp.valid=true;
	ssp.tree_mass=5;
	ssp.tree_prop=0.50;
	running=0;
	
	sps_time_valid=0;
	
	stat_diag=new StatisticsDiagram();
	stat_diag->DoShow();
	
	connect(&asetup,SIGNAL(change_area_setup(AreaSetupPars*)),
		this,SLOT(area_setup_changed(AreaSetupPars*)));
	connect(&asetup,SIGNAL(get_area_setup(AreaSetupPars*)),
		this,SLOT(fillin_area_setup(AreaSetupPars*)));
	
	/******** HEAD ********/
	QWidget *parent=this;
	QPushButton *quit_button=new QPushButton("&Quit",parent);
	connect(quit_button,SIGNAL(clicked()),this,SLOT(quit_clicked()));
	
	QPushButton *info_button=new QPushButton("Info",parent);
	connect(info_button,SIGNAL(clicked()),this,SLOT(info_clicked()));
	
	QHBoxLayout *head_layout=new QHBoxLayout();
	  head_layout->addWidget(quit_button);
	  head_layout->addSpacing(4);
	  head_layout->addWidget(info_button);
	
	int margin=5;
	int spacing=5;
	
	/******** IGNITION & CONTROL ********/
	stop_cont_button=new QPushButton("Start",this);
	  connect(stop_cont_button,SIGNAL(clicked()),this,SLOT(stop_cont_clicked()));
	  QToolTip::add(stop_cont_button,"Start / stop / continue sim.");
	
	QPushButton *reset_button=new QPushButton("&Reset",this);
	  connect(reset_button,SIGNAL(clicked()),this,SLOT(reset_clicked()));
	  QToolTip::add(reset_button,"Reset diagram.");
	
	area_button=new QPushButton("&Area...",this);
	  connect(area_button,SIGNAL(clicked()),&asetup,SLOT(map()));
	  QToolTip::add(area_button,"Specify area width and height.");
	
	ignitepars=new IgniteParsDialog(this);
	
	/******** PARAMETERS ********/
	QGroupBox *area_group=new QGroupBox("Other parameters",this);
	parent=area_group;
	
	atd=new AreaTreeDialog(0,parent);
	  connect(atd,SIGNAL(sim_setup_pars_request(SimSetupPars *)),
	    this,SLOT(sim_setup_pars_request_slot(SimSetupPars *)));
	  connect(atd,SIGNAL(sim_setup_pars_change()),
	    this,SLOT(sim_setup_pars_changed()));
	
	QString _diag_limit_str=
		"Number of data records in diagram;\n"
		"use small values (e.g. 1000) to watch\n"
		"changes while modifying settings.";
	diag_limit=new QSpinBox(0,100000,100,parent);
	  diag_limit->setSpecialValueText("[unlimited]");
	  connect(diag_limit,SIGNAL(valueChanged(int)),
	    this,SLOT(diag_limit_changed(int)));
	  QToolTip::add(diag_limit,_diag_limit_str);
	
	QWidget *wtmp;
	QHBoxLayout *diag_limit_lay=new QHBoxLayout();
	  diag_limit_lay->addWidget(wtmp=new QLabel(diag_limit,"Record &limit:",parent));
	    QToolTip::add(wtmp,_diag_limit_str);
	  diag_limit_lay->addWidget(diag_limit);
	
	QVBoxLayout *area_inplayout=new QVBoxLayout(parent,margin,spacing);
	  area_inplayout->addSpacing(15);
	  area_inplayout->addWidget(atd);
	  area_inplayout->addLayout(diag_limit_lay);
	
	/******** PROPS ********/
	
	bpropd=new BurnPropDialog(this);
	  connect(bpropd,SIGNAL(burn_prop_change(double*)),
		this,SLOT(burn_prop_change(double*)));
	  connect(bpropd,SIGNAL(burn_prop_request(double*)),
		this,SLOT(burn_prop_request_slot(double*)));
	
	/******** TOP LAYOUT ********/
	QVBoxLayout *button_layout=new QVBoxLayout();
	  button_layout->addSpacing(6);
	  button_layout->addWidget(stop_cont_button);
	  button_layout->addWidget(reset_button);
	  button_layout->addWidget(area_button);
	
	QHBoxLayout *ignite_layout=new QHBoxLayout();
	  ignite_layout->addLayout(button_layout,2);
	  ignite_layout->addSpacing(5);
	  ignite_layout->addWidget(ignitepars,5);
	
	QVBoxLayout *toplayout=new QVBoxLayout(this,/*margin=*/3,/*spacing=*/0);
	  toplayout->addLayout(head_layout);
	  toplayout->addSpacing(2);
	  toplayout->addLayout(ignite_layout);
	  toplayout->addWidget(area_group);
	  toplayout->addSpacing(5);
	  toplayout->addWidget(bpropd);
	
	startTimer(0);
}


MainWindow::~MainWindow()
{
	
}
