/*
 * statdiag.h
 * 
 * Statistics diagram. 
 * 
 * Copyright (c) 2002--2003 by Wolfgang Wieser (wwieser@gmx.de) 
 * 
 * This file may be distributed and/or modified under the terms of the 
 * GNU General Public License version 2 as published by the Free Software 
 * Foundation. 
 * 
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 */

#ifndef _BURN_STBURN_STATDIAGRAM_H_
#define _BURN_STBURN_STATDIAGRAM_H_ 1

#include <burncore/burncore.h>

#include <Qt/qwidget.h>

#include <QTX/xpainter.h>
#include <QTX/bufpointpainter.h>

#include <time.h>
#include <sys/time.h>


class BurnArea;
struct AreaSetupPars;

#define ENTRIES_PER_NODE 4096

struct LinesCache
{
	private:
		int size;
		XPoint *pnts;
		
		XPoint *head;  // where we are currently; call Reset() to reset
	
	public:
		LinesCache()
			{  size=0; pnts=NULL;  head=NULL;  }
		~LinesCache()
			{  if(pnts)  free(pnts);  pnts=NULL;  head=NULL;  size=0;  }
		
		// Call this before adding an entry. 
		inline void Reset()  {  head=pnts;  }
		// This will NOT flush. Calls Reset(). 
		// Pass number of POINTS (2 for each line). 
		void EnsureSize(int min_size);
		
		// Actually draw the lines used the passed painter; 
		// internally calls Reset(). 
		inline void DoDraw(QTX::XPainter *pnt);
		
		// Add a line to the buffer. 
		// Make sure it does not overrun. 
		inline void AddLine(int x0,int y0,int x1,int y1);
};


class StatisticsDiagram : public QWidget, QTX::XPainter
{
	public:
		struct SEntry
		{
			double tree_prop;   // used tree prop
			double dead_frac;   // ratio dead trees / initial number of trees
			int duration;       // number of cycles until the fire was over
		};
		struct EntryNode
		{
			int nent;
			SEntry ent[ENTRIES_PER_NODE];
			EntryNode *next;
		};
		struct DivEntry
		{
			double min_tree_prop;  // div from here...
			double max_tree_prop;  // ...to here (on X axis)
			size_t nent;           // number of entries counted
			double sum_dead_frac;  // sum of dead_frac
			double sq_sum_dead_frac;  // sum of dead_frac^2
			double sum_duration;   // sum of burn duration (burn cycles)
		};
	private:
		Font fnt;
		QTX::XPainter backpnt;
		QTX::XPainter framepnt;
		QTX::XPainter curvepnt;  // avg curve
		QTX::XPainter varpnt;  // variance curve
		QTX::XPainter durationpnt;  // duration curve
		QTX::BufferedPointPainter *ppnt;
		LinesCache avg_lines;
		LinesCache dur_lines;
		
		EntryNode *first,*last;
		size_t nent;  // total number of entries
		size_t entry_limit;   // 0 -> unlimited
		
		size_t change_ent;  // only needed for display 
		
		// Where the old entries are overwritten by new ones 
		// if entry_limit is reached: 
		EntryNode *el_write_node;
		SEntry *el_write_pos;
		
		// NOTE: these cannot be efficiently maintained if 
		// entry_limit is switched on and thus remain NULL 
		// in this case. 
		SEntry *duration_max;
		SEntry *dead_frac_max;
		SEntry *dead_frac_min;
		
		struct timeval last_curve_update;
		
		double sims_per_sec;
		double sps_accu;
		int sps_accu_nent;
		
		int diag_x0;
		int diag_y0;
		int diag_width;
		int diag_height;
		
		int A_n_div_entries;
		DivEntry *A_div_ent;
		int B_n_div_entries;
		DivEntry *B_div_ent;
		
		//  Window already shown? (On init only.) 
		int ismapped;
		
		bool do_draw_dots;       // draw a dot for each tree
		int do_draw_avg_curve;   // 0 -> no; 1 -> avg; 2 -> avg and var
		int do_draw_duration_curve;
		
		inline int Prop2DiagX(double p)
			{  return(diag_x0+int(p*diag_width + 0.5));  }
		inline int Prop2DiagY(double p)
			{  return(diag_y0 + (diag_height - int(p*diag_height + 0.5)));  }
		
		SEntry *_AllocNextEntry();
		inline void _UpdateMinMaxPtrs(SEntry *se);
		
		inline void _RelimitAddRecord(SEntry *data);
		
		void _ResetDivArray(int n_div_entries,DivEntry *div_ent);
		void _AddDivArrayEntry(int n_div_entries,DivEntry *div_ent,SEntry *se);
		void _DelDivArrayEntry(int n_div_entries,DivEntry *div_ent,SEntry *se);
		
		void _RecalcSize(int nw,int nh);
		
		void _RedrawAll();
		void _DrawDiagramLines(int dir,int ndiv);
		void _DrawNEntText(bool all,bool do_nent,bool do_sps);
		void _DrawAddEntry(SEntry *se,SEntry *undraw_se,bool redraw_sps);
		
		void _HandleKeyEvent(QKeyEvent *ev,int stat);
		
		void paintEvent(QPaintEvent *);
		void resizeEvent(QResizeEvent *);
		void keyPressEvent(QKeyEvent *ev);
		void keyReleaseEvent(QKeyEvent *ev);
	public:
		StatisticsDiagram(QWidget *parent=NULL,const char *name=NULL,WFlags f=0);
		~StatisticsDiagram();
		
		// Does no reset internally. 
		void area_setup_changed(AreaSetupPars *spar);
		
		void DoShow()
			{  ismapped=1;  show();  }
		
		void Reset();
		
		void ReDraw()
			{  _RedrawAll();  }
		
		size_t GetEntryLimit()
			{  return(entry_limit);  }
		void ChangeEntryLimit(size_t new_limit);
		
		void AddRecord(double tree_prop,BurnArea *a,int iterations,long usec=-1);
		
		void GetDivArrayA(int *ret_size,DivEntry **ret_ents)
			{  *ret_size=A_n_div_entries;  *ret_ents=A_div_ent;  }
		void GetDivArrayB(int *ret_size,DivEntry **ret_ents)
			{  *ret_size=B_n_div_entries;  *ret_ents=B_div_ent;  }
};

#endif  /* _BURN_STBURN_STATDIAGRAM_H_ */
